package de.hamster.simulation.model;

import java.util.ArrayList;
import java.util.Observable;

import de.hamster.model.CreateInstruction;
import de.hamster.model.HamsterException;
import de.hamster.model.HamsterInitialisierungsException;
import de.hamster.model.HamsterInstruction;
import de.hamster.model.Instruction;
import de.hamster.model.InstructionProcessor;
import de.hamster.model.KachelLeerException;
import de.hamster.model.MauerDaException;
import de.hamster.model.MaulLeerException;
import de.hamster.model.TerminalInstruction;
import de.hamster.model.TerrainCellInstruction;
import de.hamster.model.TerrainInstruction;

/**
 * Dies ist das Model der Simulationskomponente. Es implementiert das
 * InstructionProcessor Interface, so dass es die Befehle von Hamsterprogrammen
 * ausfuehren kann. Die verschiedenen View-Komponenten melden sich ueber die
 * Observer-Schnittstelle an dieser Komponente an, so dass sie die jeweiligen
 * Veraenderungen darstellen koennen. Ausserdem koennen sich diverse LogSinks an
 * am SimulationModel anmelden, die entsprechende Logausgaben darstellen.
 * 
 * @author $Author: djasper $
 * @version $Revision: 1.3 $
 */
public class SimulationModel extends Observable implements InstructionProcessor {
	static final int[] DX = {0, 1, 0, -1};
	static final int[] DY = {-1, 0, 1, 0};

	/**
	 * This argument notifies observers of a state change.
	 */
	public static final String STATE = "state change";
	/**
	 * This argument notifies observers of a change of the terrain.
	 */
	public static final String TERRAIN = "terrain";

	/**
	 * In this state, no simulation is running.
	 */
	public static final int NOT_RUNNING = 0;

	/**
	 * In this state, a simulation is running.
	 */
	public static final int RUNNING = 1;

	/**
	 * In this state, a simulation is running, but suspended.
	 */
	public static final int SUSPENDED = 2;

	/**
	 * This array contains hamsters, that where constructed during the current
	 * simulation.
	 */
	private ArrayList hamster;

	/**
	 * This is the current terrain.
	 */
	private Terrain terrain;
	
	protected Terrain savedTerrain;

	/**
	 * This is the current state of the simulation.
	 */
	private int state;

	private Terminal terminal;

	private ArrayList logSinks;

	public SimulationModel() {
		hamster = new ArrayList();
		terrain = new Terrain(16, 6);
		logSinks = new ArrayList();
	}

	public void addLogSink(LogSink sink) {
		logSinks.add(sink);
	}

	public void log(LogEntry entry) {
		for (int i = 0; i < logSinks.size(); i++) {
			((LogSink) logSinks.get(i)).logEntry(entry);
		}
	}
	public void clearLogs() {
		for (int i = 0; i < logSinks.size(); i++) {
			((LogSink) logSinks.get(i)).clearLog();
		}
	}

	public Terrain getTerrain() {
		return terrain;
	}

	public ArrayList getHamster() {
		return hamster;
	}

	public void setTerrain(Terrain terrain) {
		this.terrain = terrain;
		hamster = new ArrayList();
		setChanged();
		notifyObservers(TERRAIN);
	}

	public boolean containsHamster(int x, int y) {
		Hamster h = terrain.getDefaultHamster();
		if (h.getX() == x && h.getY() == y)
			return true;
		for (int i = 0; i < hamster.size(); i++) {
			h = (Hamster) hamster.get(i);
			if (h.getX() == x && h.getY() == y)
				return true;
		}
		return false;
	}

	public int getState() {
		return state;
	}

	public void removeHamster() {
		hamster.removeAll(hamster);
		setChanged();
		notifyObservers(TERRAIN);
	}

	public void start() {
		removeHamster();
		clearLogs();
		savedTerrain = new Terrain(terrain);
	}
	
	public void reset() {
		if(savedTerrain == null)
			return;
		terrain = savedTerrain;
		hamster = new ArrayList();
		setChanged();
		notifyObservers(TERRAIN);
		savedTerrain = null;
	}
	
	public void finished() {
		setState(SimulationModel.NOT_RUNNING);
	}

	public Object processHamsterInstruction(HamsterInstruction instruction) {
		int id = instruction.getHamster();
		if (instruction instanceof CreateInstruction) {
			CreateInstruction ci = (CreateInstruction) instruction;
			if (terrain.getWall(ci.getCol(), ci.getRow()) || ci.getDir() < 0
					|| ci.getDir() > 3 || ci.getMouth() < 0)
				throw new HamsterInitialisierungsException(null);
			hamster.add(new Hamster(id, ci.getCol(), ci.getRow(), ci.getDir(),
					ci.getMouth()));
			setChanged();
			notifyObservers(TERRAIN);
			return "ok";
		} else if (instruction instanceof TerminalInstruction) {
			TerminalInstruction ti = (TerminalInstruction) instruction;
			return "ok";
		} else {
			switch (instruction.getType()) {
				case HamsterInstruction.FORWARD :
					return forward(id);
				case HamsterInstruction.TURN_LEFT :
					return turnLeft(id);
				case HamsterInstruction.LAY_DOWN :
					return layDown(id);
				case HamsterInstruction.PICK_UP :
					return pickUp(id);

				case HamsterInstruction.FREE :
					return new Boolean(free(id));
				case HamsterInstruction.CORN_AVAILABLE :
					return new Boolean(cornAvailable(id));
				case HamsterInstruction.MOUTH_EMPTY :
					return new Boolean(mouthEmpty(id));

				case HamsterInstruction.GET_ROW :
					return new Integer(getHamster(id).getY());
				case HamsterInstruction.GET_COL :
					return new Integer(getHamster(id).getX());
				case HamsterInstruction.GET_DIR :
					return new Integer(getHamster(id).getDir());
				case HamsterInstruction.GET_MOUTH :
					return new Integer(getHamster(id).getMouth());

				case HamsterInstruction.GET_DATA :
					return new int[]{getHamster(id).getY(),
							getHamster(id).getX(), getHamster(id).getDir(),
							getHamster(id).getMouth()};
			}
		}
		return null;
	}
	public Object processTerrainCellInstruction(TerrainCellInstruction tci) {
		switch (tci.getType()) {
			case TerrainCellInstruction.IS_WALL :
				return new Boolean(terrain.getWall(tci.getCol(), tci.getRow()));
			case TerrainCellInstruction.COUNT_CORN :
				return new Integer(terrain.getCornCount(tci.getCol(), tci
						.getRow()));
			case TerrainCellInstruction.COUNT_HAMSTER :
				return new Integer(getHamsterIDs(tci.getCol(), tci.getRow())
						.size());
			case TerrainCellInstruction.GET_HAMSTER :
				return getHamsterIDs(tci.getCol(), tci.getRow());
		}
		return null;
	}

	public Object processTerrainInstruction(TerrainInstruction instruction) {
		switch (instruction.getType()) {
			case TerrainInstruction.COUNT_ROWS :
				return new Integer(terrain.getHeight());
			case TerrainInstruction.COUNT_COLS :
				return new Integer(terrain.getWidth());
			case TerrainInstruction.COUNT_CORN :
				return new Integer(countTotalCorn());
			case TerrainInstruction.COUNT_HAMSTER :
				return new Integer(hamster.size() + 1);
			case TerrainInstruction.GET_HAMSTER :
				return getHamsterIDs();
		}
		return null;
	}

	public Object processTerminalInstruction(TerminalInstruction instruction) {
		int id = instruction.getHamster();
		String message = instruction.getMessage();
		switch (instruction.getType()) {
			case TerminalInstruction.WRITE :
				terminal.write(id, message);
				return "ok";
			case TerminalInstruction.READ_INT :
				return new Integer(terminal.readInt(id, message));
			case TerminalInstruction.READ_STRING :
				return terminal.readString(id, message);
		}
		return null;
	}

	public Object process(Instruction instruction) {
		try {
			Object result = null;
			if (instruction instanceof TerminalInstruction) {
				result = processTerminalInstruction((TerminalInstruction) instruction);
			} else if (instruction instanceof HamsterInstruction) {
				result = processHamsterInstruction((HamsterInstruction) instruction);
			} else if (instruction instanceof TerrainCellInstruction) {
				result = processTerrainCellInstruction((TerrainCellInstruction) instruction);
			} else if (instruction instanceof TerrainInstruction) {
				result = processTerrainInstruction((TerrainInstruction) instruction);
			}
			log(new LogEntry(instruction, result));
			return result;
		} catch (HamsterException e) {
			log(new LogEntry(instruction, e));
			return e;
		}
	}

	public ArrayList getHamsterIDs() {
		ArrayList ids = new ArrayList();
		for (int i = 0; i < hamster.size(); i++) {
			Hamster h = (Hamster) hamster.get(i);
			ids.add(new Integer(h.getId()));
		}
		ids.add(new Integer(-1));
		return ids;
	}

	public ArrayList getHamsterIDs(int col, int row) {
		ArrayList ids = new ArrayList();
		for (int i = 0; i < hamster.size(); i++) {
			Hamster h = (Hamster) hamster.get(i);
			if (h.getX() != col)
				continue;
			if (h.getY() != row)
				continue;
			ids.add(new Integer(h.getId()));
		}
		Hamster h = terrain.getDefaultHamster();
		if (h.getX() == col && h.getY() == row)
			ids.add(new Integer(-1));
		return ids;
	}

	public int countTotalCorn() {
		int total = 0;
		for (int i = 0; i < terrain.getWidth(); i++) {
			for (int j = 0; j < terrain.getHeight(); j++) {
				total += terrain.getCornCount(i, j);
			}
		}
		return total;
	}

	public void processException(Throwable t) {
		terminal.write(t);
	}

	private Hamster getHamster(int id) {
		if (id == -1)
			return terrain.getDefaultHamster();
		for (int i = 0; i < hamster.size(); i++) {
			Hamster h = (Hamster) hamster.get(i);
			if (h.getId() == id)
				return h;
		}
		return null;
	}

	public Object turnLeft(int id) {
		Hamster hamster = getHamster(id);
		hamster.setDir((hamster.getDir() + 3) % 4);
		setChanged();
		notifyObservers(TERRAIN);
		return "ok";
	}

	private Object forward(int id) {
		Hamster hamster = getHamster(id);
		int nextX = hamster.getX() + DX[hamster.getDir()];
		int nextY = hamster.getY() + DY[hamster.getDir()];
		if (terrain.getWall(nextX, nextY)) {
			throw new MauerDaException(null, nextY, nextX);
		} else {
			hamster.setXY(nextX, nextY);
			setChanged();
			notifyObservers(TERRAIN);
			return "ok";
		}
	}

	private Object pickUp(int id) {
		Hamster hamster = getHamster(id);
		int count = terrain.getCornCount(hamster.getX(), hamster.getY());
		if (count == 0) {
			throw new KachelLeerException(null, hamster.getY(), hamster.getX());
		} else {
			hamster.setMouth(hamster.getMouth() + 1);
			terrain.setCornCount(hamster.getX(), hamster.getY(), count - 1);
			setChanged();
			notifyObservers(TERRAIN);
			return "ok";
		}
	}

	private Object layDown(int id) {
		Hamster hamster = getHamster(id);
		int count = terrain.getCornCount(hamster.getX(), hamster.getY());
		if (hamster.getMouth() == 0) {
			throw new MaulLeerException(null);
		} else {
			hamster.setMouth(hamster.getMouth() - 1);
			terrain.setCornCount(hamster.getX(), hamster.getY(), count + 1);
			setChanged();
			notifyObservers(TERRAIN);
			return "ok";
		}
	}

	private boolean free(int id) {
		Hamster hamster = getHamster(id);
		return !terrain.getWall(hamster.getX() + DX[hamster.getDir()], hamster
				.getY()
				+ DY[hamster.getDir()]);
	}

	private boolean cornAvailable(int id) {
		Hamster hamster = getHamster(id);
		return terrain.getCornCount(hamster.getX(), hamster.getY()) != 0;
	}

	private boolean mouthEmpty(int id) {
		Hamster hamster = getHamster(id);
		return hamster.getMouth() == 0;
	}

	private void setState(int state) {
		this.state = state;
		setChanged();
		notifyObservers(STATE);
	}

	/**
	 * @param terminal
	 *            The terminal to set.
	 */
	public void setTerminal(Terminal terminal) {
		this.terminal = terminal;
	}
}